#include <QWebSettings>
#include <QSqlDatabase>
#include <QSettings>
#include <QShortcut>
#include <QEventLoop>
#include <QDesktopWidget>
#include <QWebFrame>
#include <QWebElement>

#include <QDebug>

#include "ui/exitpassworddialog.h"
#include "ui/widgets/browser/browser.h"
#include "ui/widgets/webpage.h"
#include "viewer.h"
#include "utils.h"

bool PlaylistEventFilter::eventFilter(QObject *obj, QEvent *event)
{
    if (m_paused)
        return false;

    switch (event->type())
    {
        case QEvent::MouseMove:
        case QEvent::MouseButtonPress:
        case QEvent::KeyPress:
            qobject_cast<QWidget *>(obj)->setCursor(Qt::BlankCursor);
            qobject_cast<Viewer *>(obj->parent())->emitMouseEvent(); /* hum, not nice... */
            return true;
        default:
            return false;
    }
}

bool BrowserEventFilter::eventFilter(QObject *obj, QEvent *event)
{
    QMouseEvent *me = dynamic_cast<QMouseEvent *>(event);
    QInputEvent *ie = dynamic_cast<QInputEvent *>(event);
    if (me || ie)
        qobject_cast<Viewer *>(obj->parent())->resetTimer();

    return false;
}

Viewer::Viewer()
    : QStackedWidget()
{
    settings = Settings::load();
    prevWidget = 0;

    connect(&m_timer, SIGNAL(timeout()), this, SLOT(onTimerTimeout()));

    m_playlistEventFilter = new PlaylistEventFilter;
    m_browserEventFilter = new BrowserEventFilter;
    m_unlockTimer.setSingleShot(true);
    connect(&m_unlockTimer, SIGNAL(timeout()),
            this, SLOT(unlockPlaylistEventFilter()));

    loadWindowSettings();
    setShortcuts();

    if (settings->get("main/fullscreen").toBool())
        showFullScreen();

    QWebSettings *websettings = QWebSettings::globalSettings();
    websettings->setAttribute(QWebSettings::JavascriptCanOpenWindows, true);
    websettings->setAttribute(QWebSettings::PluginsEnabled, true);
    websettings->setAttribute(QWebSettings::JavaEnabled, true);

    QString fontStr = settings->get("webpage/default_font").toString();
    if (!fontStr.isEmpty())
    {
        QFont font(fontStr);
        websettings->setFontFamily(QWebSettings::StandardFont, font.family());
        websettings->setFontSize(QWebSettings::DefaultFontSize, font.pointSize());
    }

    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE");
    db.setDatabaseName(settings->cookiesFile());
    db.open();
}

void Viewer::emitMouseEvent()
{
    emit(mouseEvent());
}

bool Viewer::isReadyToShoot()
{
    if(prevWidget->inherits(Webpage::staticMetaObject.className())){
        return qobject_cast<Webpage *>(prevWidget)->isLoaded()&&readyToShoot;
    }
    return readyToShoot;
}

void Viewer::loadWindowSettings()
{
    QSettings appsettings;
    appsettings.beginGroup("Viewer");
    resize(appsettings.value("size", QSize(800, 600)).toSize());
    move(appsettings.value("pos", QPoint(200, 200)).toPoint());
    appsettings.endGroup();
}

void Viewer::saveWindowSettings()
{
    QSettings appsettings;
    appsettings.beginGroup("Viewer");
    appsettings.setValue("size", size());
    appsettings.setValue("pos", pos());
    appsettings.endGroup();
}

void Viewer::setShortcuts()
{
    QShortcut *shortcut;

    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q), this);
    shortcut->setContext(Qt::ApplicationShortcut);
    connect(shortcut, SIGNAL(activated()), this, SLOT(bye()));

    shortcut = new QShortcut(QKeySequence(Qt::Key_Escape), this);
    shortcut->setContext(Qt::ApplicationShortcut);
    connect(shortcut, SIGNAL(activated()), this, SLOT(bye()));
}

void Viewer::bye()
{
    if (settings->get("main/admin_passwd").toBool())
    {
        ExitPasswordDialog d;
        QString password = d.run();
        if (password.isEmpty() || ! Utils::validateAdminPassword(password))
            return;
    }

    if (prevWidget)
    {
        Runnable *r = dynamic_cast<Runnable *>(prevWidget);
        if (r)
            r->stop();
    }
    saveWindowSettings();
    destroy();
    exit(0);
}

void Viewer::onTimerTimeout()
{
    m_timer.stop();
    emit(end());
}

void Viewer::stop()
{
    m_timer.stop();
}

void Viewer::setTimer()
{
    m_timer.setInterval(settings->get("main/stop_session_after_seconds").toInt() * 1000);
    m_timer.start();
}

void Viewer::resetTimer()
{
    m_timer.start();
}

void Viewer::closeEvent(QCloseEvent *event)
{
    bye();
    event->ignore();
}

void Viewer::next(QWidget *widget)
{
    readyToShoot = false;
    addWidget(widget);
    widget->setMouseTracking(true);

    Browser *b = qobject_cast<Browser *>(widget);
    if (b)
    {
        if (settings->get("main/stop_session_after_seconds").toInt())
        {
            widget->installEventFilter(m_browserEventFilter);
            connect(widget, SIGNAL(mouseMoved()), this, SLOT(resetTimer()));
            setTimer();
        }
    }
    else
    {
        widget->setCursor(Qt::BlankCursor);
        if (settings->get("slideshow/stop_on_event").toBool())
            widget->installEventFilter(m_playlistEventFilter);
    }

    connect(widget, SIGNAL(end()), this, SLOT(onWidgetTimeout()));



    /* VLC send a MouseMove event that might stop the slideshow, so temporarily
     * ignore the events
     */
    m_playlistEventFilter->setPaused(true);
    m_unlockTimer.start(1000);
    dynamic_cast<Runnable *>(widget)->run();
    if (prevWidget)
    {
        removeWidget(prevWidget);
        prevWidget->deleteLater();
    }
    prevWidget = widget;
    if(widget->inherits(Webpage::staticMetaObject.className())){
        QEventLoop loop;
        Webpage * page = qobject_cast<Webpage *>(widget);
        connect(page,SIGNAL(loadFinished(bool)),&loop,SLOT(quit()));
        loop.exec();
    }

    widget->show();
    setCurrentWidget(widget);
    readyToShoot = true;
}

void Viewer::pause()
{
    Runnable *r = dynamic_cast<Runnable *>(prevWidget);
    if (r)
        r->pause();
}

void Viewer::showText(const QString &text, int duration)
{
    readyToShoot = false;
    qDebug()<<QUrl::fromLocalFile("/usr/local/share/linutop-webkiosk/text.html").toString();
    PlaylistItem * item = new PlaylistItem("file:///usr/local/share/linutop-webkiosk/text.html",PlaylistItem::TypeWebpage,duration);
    Webpage * page = new Webpage(item,this);
    page->run();
    addWidget(page);
    QEventLoop loop;
    connect(page,SIGNAL(loadFinished(bool)),&loop,SLOT(quit()));
    loop.exec();
    page->page()->mainFrame()->evaluateJavaScript(QString("document.getElementById('text').innerHTML = '%1';").arg(text));
    qDebug()<<page->page()->mainFrame()->toHtml();
    page->show();
    setCurrentWidget(page);
    this->pause();
    readyToShoot = true;
}

void Viewer::unlockPlaylistEventFilter()
{
    m_playlistEventFilter->setPaused(false);
}

void Viewer::onWidgetTimeout()
{
    m_timer.stop();
    emit(end());
}
